/**********************************************************************
 * FILE: ama_scan.c
 * AUTHOR: Fabian Buske / Robert McLeay for refactoring
 * PROJECT: MEME
 * COPYRIGHT: 2007-2008, UQ
 * VERSION: $Revision: 1.0$
 * DESCRIPTION: Routines to perform average motif affinity scans
 *
 **********************************************************************/

#include "ama_scan.h"

char* motif_name = "motif";

/*************************************************************************
 * Converts the motif frequency matrix into a odds matrix
 *************************************************************************/
void convert_to_odds_matrix(MOTIF_T* motif, ARRAY_T* bg_freqs){
	const int alph_size = get_alph_size(ALPH_SIZE);
	int motif_position_index,alph_index;

	const int num_motif_positions = get_num_rows(motif->freqs);
	for (alph_index=0;alph_index<alph_size;++alph_index){
		double bg_likelihood = get_array_item(alph_index, bg_freqs);
		for (motif_position_index=0;motif_position_index<num_motif_positions;++motif_position_index){
			motif->freqs->rows[motif_position_index]->items[alph_index] /= bg_likelihood;
		}
	}
}

/*************************************************************************
 * Calculate the odds score for each motif-sized window at each
 * site in the sequence using the given nucleotide frequencies.
 *
 * This function is a lightweight version based on the one contained in
 * motiph-scoring. Several calculations that are unnecessary for gomo
 * have been removed in order to speed up the process
 *************************************************************************/
double score_sequence(
		SEQ_T*        seq,				// sequence to scan (IN)
		MOTIF_T*      motif,				// motif already converted to odds values (IN)
		char*         feature_type,		// motif name (IN)
		char*         seq_name,			// sequence name (IN)
		int 			method,				// method used for scoring (IN)
		SCANNED_SEQUENCE_T* scanned_seq	// the scanned sequence results (OUT)
		)
{
	assert(seq != NULL);
	assert(motif != NULL);

	// Get data for sequence
	if (seq_name == NULL) {
		seq_name = get_seq_name(seq);
	}
	char* raw_seq = get_raw_sequence(seq);
	int seq_length = get_seq_length(seq);

	// Prepare storage for the string representing the portion
	// of the reference sequence within the window.
	char* window_seq = (char *) mm_malloc(sizeof(char) * (motif->length + 1));
	window_seq[motif->length] = '\0';

	// Get a character for the strand.
	char strand = 0;
	if (motif->id!= NULL) {
		strand = *motif->id; /* FIXME: This will not work when we generalize
					the motif naming convention. */
	}
	else {
		strand = '.';
	}

	int max_index = seq_length - motif->length;
	char* alphabet = get_alphabet(FALSE);
	int alph_size = get_alph_size(ALPH_SIZE);
	double* odds =  (double*) mm_malloc(sizeof(double)*max_index);

	// For each site in the sequence
	int seq_index;
	for (seq_index = 0; seq_index < max_index; seq_index++) {
		BOOLEAN_T skip_window = FALSE;
		BOOLEAN_T found_gap = FALSE;
		BOOLEAN_T found_ambiguity = FALSE;
		double odd = 1.0;

		// For each site in the motif window
		int motif_position;
		for (motif_position = 0; motif_position < motif->length; motif_position++) {
			char c = raw_seq[seq_index + motif_position];
			window_seq[motif_position] = c;

			// Check for gaps at this site
			if(c == '-' || c == '.') {
				found_gap = TRUE;
				skip_window = TRUE;
				break;
			}

			// Check for ambiguity codes at this site
			int alph_index = alphabet_index(c, alphabet);
			if (alph_index > alph_size) {
				found_ambiguity = TRUE;
				skip_window = TRUE;
				break;
			}
			odd *= get_matrix_cell(motif_position, alph_index, motif->freqs);
		}
		odds[seq_index] = odd;
	}

	// return odds as requested (MAX or AVG scoring)
	double requested_odds = 0.0;
	if (method == AVG_ODDS){
		for (seq_index = 0; seq_index < max_index; seq_index++) {
			requested_odds += odds[seq_index];
		}
		requested_odds /= max_index;
	} else if (method == MAX_ODDS){
		for (seq_index = 0; seq_index < max_index; seq_index++) {
			if (odds[seq_index] > requested_odds){
				requested_odds = odds[seq_index];
			}
		}
	}

	myfree(odds);
	myfree(window_seq);
	return requested_odds;
}

/**********************************************************************
  ama_sequence_scan()

  scan a given sequence with a specified motif using either
  average motif affinity scoring or maximum one. In addition z-scores
  may be calculated.

  The motif has to be converted to odds in advance (in order
  to speed up the scanning). Use convert_to_odds_matrix() once for each
  motif.

  The result will be stored in the scanned_sequence parameter
 **********************************************************************/
void ama_sequence_scan(
		SEQ_T* sequence,	// the sequence to scan INPUT
		MOTIF_T* motif,		// the motif to scan with (converted to odds matrix) INPUT
		MOTIF_T* rev_motif, // the reversed motif
		int scoring,		// the scoring function to apply AVG_ODDS or MAX_ODDS INPUT
		int zscoring,		// the number of shuffled sequences used for z-score computation INPUT
		SCANNED_SEQUENCE_T* scanned_seq,	// where the results are put
		BOOLEAN_T scan_both_strands			//Should we scan with both motifs and combine scores
		){
	assert(zscoring >= 0);
	char* seq_name = get_seq_name(sequence);

	// Score the forward strand.
	double odds = score_sequence(
			sequence,
			motif,
			motif_name,
			seq_name,
			scoring,
			scanned_seq
			);

	// Score the reverse strand.
	if (scan_both_strands) {

		double rev_odds = score_sequence(
				sequence,
				rev_motif,
				motif_name,
				seq_name,
				scoring,
				scanned_seq
				);

		if (scoring == AVG_ODDS){
			odds = (odds+rev_odds)/2.0;
		} else if (scoring == MAX_ODDS){
			odds = max(odds,rev_odds);
		}

	}
	set_scanned_sequence_score(scanned_seq, odds);

	// compute z score
	if (zscoring>0){
		// Create a temp pattern
		PATTERN_T *shuffle_pattern = allocate_pattern("shuffled", "shuffled");

		double* shuffle_scores = mm_malloc(sizeof(double)*zscoring);

		/* create the suffled sequences*/
		int i;
		int seed = 0;
		for (i=0;i<zscoring;++i){
			SEQ_T* shuffled_s =  NULL;
			shuffle_sequence(sequence,seed+i,&shuffled_s);

			char* suffled_seq_name = "shuffled sequences";

			// Create a scanned_sequence record and record it in pattern.
			SCANNED_SEQUENCE_T *shuffled_random_seq =
				allocate_scanned_sequence(suffled_seq_name, suffled_seq_name, shuffle_pattern);
			set_scanned_sequence_length(shuffled_random_seq, get_seq_length(shuffled_s));

			// Score the forward strand.
			double z_odds = score_sequence(
					shuffled_s,
					motif,
					get_motif_id(motif),
					suffled_seq_name,
					scoring,
					shuffled_random_seq
					);

			// Score the reverse strand.
			if (scan_both_strands) {
				double rev_z_odds = score_sequence(
						shuffled_s,
						rev_motif,
						get_motif_id(motif),
						suffled_seq_name,
						scoring,
						shuffled_random_seq
						);
				if (scoring == AVG_ODDS){
					z_odds = (z_odds+rev_z_odds)/2.0;
				} else if (scoring == MAX_ODDS){
					z_odds = max(z_odds,rev_z_odds);
				}
			}
			free_seq(shuffled_s);
			shuffle_scores[i] = z_odds;
		} // sequence shuffling

		ZSCORE_T *z = allocate_zscore(shuffle_scores, zscoring);
		double zscore = get_zscore_z(z,odds);
		if (verbosity >= HIGH_VERBOSE) {
			fprintf(stderr, "Motif %s Sequence %s transformed score %g to z-score %g.\n",get_motif_id(motif), get_scanned_sequence_accession(scanned_seq),odds,zscore);
		}
		set_scanned_sequence_score(scanned_seq, zscore);

		/* free shuffle related stuff since its not required anymore */
		free_pattern(shuffle_pattern);
		myfree(shuffle_scores);
		destroy_zscore(z);
	} // z score computation
}
