/********************************************************************
 * FILE: fimo.c
 * AUTHOR: William Stafford Noble, Charles E. Grant
 * CREATE DATE: 12/17/2004
 * PROJECT: MEME suite
 * COPYRIGHT: 2004-2007, UW
 ********************************************************************/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "matrix.h"
#include "alphabet.h"
#include "cisml.h"
#include "dir.h"
#include "fasta-io.h"
#include "hash_alph.h"
#include "io.h"
#include "meme-io.h"
#include "motiph-scoring.h"
#include "string-list.h"
#include "simple-getopt.h"
#include "utils.h"
#include "xml-util.h"

#define MAX_PALINDROMIC_OVERLAP_ALLOWED 0.8
#define min(a,b)      (a<b)?a:b

#define HTML_STYLESHEET "fimo-to-html.xsl"
#define CSS_STYLESHEET "cisml.css"
#define TEXT_STYLESHEET "fimo-to-text.xsl"
#define GFF_STYLESHEET "fimo-to-gff.xsl"
#define XML_FILENAME "fimo.xml"
#define HTML_FILENAME "fimo.html"
#define TEXT_FILENAME "fimo.txt"
#define GFF_FILENAME "fimo.gff"

char* program_name = NULL;
VERBOSE_T verbosity = NORMAL_VERBOSE;

/**********************************************************************
 * This function saves the FIMO results as a set of files in a 
 * directory. The files include:
 *   fimo.xml which uses the CisML schema
 *   fimo.txt which uses plain text with tab-delimited columns
 *   fimo.gff which uses the gff file format.
 *********************************************************************/
static void output_full_results(
  CISML_T *cisml, 
  char *output_dirname, 
  BOOLEAN_T allow_clobber
) {

  // Create output directory
  if (create_output_directory(
       output_dirname, 
       allow_clobber, 
       FALSE /* Don't print warning messages */
      )
    ) {
    // Failed to create output directory.
    die("Couldn't create output directory %s.\n", output_dirname);
  }
  // Create the paths to the output files
  char *html_stylesheet_path = make_path_to_file(ETC_DIR, HTML_STYLESHEET);
  char *css_stylesheet_path = make_path_to_file(ETC_DIR, CSS_STYLESHEET);
  char *text_stylesheet_path = make_path_to_file(ETC_DIR, TEXT_STYLESHEET);
  char *gff_stylesheet_path = make_path_to_file(ETC_DIR, GFF_STYLESHEET);
  char *xml_path = make_path_to_file(output_dirname, XML_FILENAME);
  char *html_path = make_path_to_file(output_dirname, HTML_FILENAME);
  char *text_path = make_path_to_file(output_dirname, TEXT_FILENAME);
  char *gff_path = make_path_to_file(output_dirname, GFF_FILENAME);
  char *html_stylesheet_copy_path = make_path_to_file(output_dirname, HTML_STYLESHEET);
  char *css_stylesheet_copy_path = make_path_to_file(output_dirname, CSS_STYLESHEET);
  FILE *xml_file = fopen(xml_path, "w");
  if (!xml_file) {
    die("Couldn't open file %s for output.\n", xml_path);
  }

  // Output XML 
  print_cisml(xml_file, cisml, TRUE, HTML_STYLESHEET, TRUE);
  fclose(xml_file);

  // Output HTML
  print_xml_filename_to_filename_using_stylesheet(xml_path, html_stylesheet_path, html_path);

  // Output text
  print_xml_filename_to_filename_using_stylesheet(xml_path, text_stylesheet_path, text_path);

  // Output GFF
  print_xml_filename_to_filename_using_stylesheet(xml_path, gff_stylesheet_path, gff_path);

  // Copy XML to HTML and CSS stylesheets to output directory
  copy_file(html_stylesheet_path, html_stylesheet_copy_path);
  copy_file(css_stylesheet_path, css_stylesheet_copy_path);

  myfree(html_stylesheet_path);
  myfree(html_stylesheet_copy_path);
  myfree(css_stylesheet_path);
  myfree(css_stylesheet_copy_path);
  myfree(text_stylesheet_path);
  myfree(gff_stylesheet_path);
  myfree(xml_path);
  myfree(html_path);
  myfree(text_path);
  myfree(gff_path);
}

/**********************************************************************
 * This function saves the FIMO results as a set of files in a 
 * directory. The files include:
 *   fimo.xml which uses the CisML schema
 *   fimo.txt which uses plain text with tab-delimited columns
 *   fimo.gff which uses the gff file format.
 *********************************************************************/
static void output_only_text_results(CISML_T *cisml) {

  const char* etc_dir = get_meme_etc_dir();

  // Create the paths to the output files
  char *text_stylesheet_path = make_path_to_file(etc_dir, TEXT_STYLESHEET);
  char tmp_filename[L_tmpnam]; 
  char *filename = tmpnam(tmp_filename);
  if (! filename) {
    die("Couldn't create temporary file for text results\n");
  }
  FILE *xml_file = fopen(filename, "w");
  if (!xml_file) {
    die("Couldn't open file %s for output.\n", "fimo.xml");
  }

  // Output XML 
  print_cisml(xml_file, cisml, TRUE, HTML_STYLESHEET, TRUE);
  fclose(xml_file);

  // Output text
  print_xml_filename_to_file_using_stylesheet(filename, text_stylesheet_path, stdout);

  myfree(text_stylesheet_path);
  remove(filename);

}

/*************************************************************************
 * Entry point for fimo
 *************************************************************************/
int main(int argc, char *argv[]) {

  char* bg_filename = NULL;
  char *output_dirname = "fimo_out";
  BOOLEAN_T allow_clobber = TRUE;
  int max_seq_length = MAX_SEQ;
  char* motif_name = "motif"; // Use this motif name in the output.
  char* seq_name = NULL;      // Use this sequence name in the output.
  STRING_LIST_T* selected_motifs = NULL;
  double pthresh = 1e-4;
  double qthresh = 1.0;
  double pseudocount = 0.1;
  BOOLEAN_T text_only = FALSE;
  BOOLEAN_T scan_both_strands = TRUE;
  BOOLEAN_T compute_qvalues = TRUE;
  program_name = "fimo";

  /**********************************************
   * COMMAND LINE PROCESSING
   **********************************************/

  const int num_options = 11;
  cmdoption const motif_scan_options[] = {
    {"bgfile", REQUIRED_VALUE},
    {"max-seq-length", REQUIRED_VALUE},
    {"motif", REQUIRED_VALUE},
    {"motif-pseudo", REQUIRED_VALUE},
    {"norc", NO_VALUE},
    {"o", REQUIRED_VALUE},
    {"oc", REQUIRED_VALUE},
    {"output-pthresh", REQUIRED_VALUE},
    {"output-qthresh", REQUIRED_VALUE},
    {"text", NO_VALUE},
    {"verbosity", REQUIRED_VALUE}
  };

  int option_index = 0;

  // Define the usage message.
  char usage[] = 
    "USAGE: fimo [options] <motif file> <sequence file>\n"
    "\n"
    "   Options:\n"
    "     --bgfile <background> (default from NR sequence database)\n"
    "     --max-seq-length <int> (default=1e6)\n"
    "     --motif <id> (default=all)\n"
    "     --motif-pseudo <float> (default=0.1)\n"
    "     --norc\n"
    "     --o <output dir> (default=fimo_out)\n" 
    "     --oc <output dir> (default=fimo_out)\n" 
    "     --output-pthresh <float> (default 1e-4)\n"
    "     --output-qthresh <float> (default 1.0)\n"
    "     --text\n"
    "     --verbosity [1|2|3|4] (default 2)\n"
    "\n";

  // Parse the command line.
  if (simple_setopt(argc, argv, num_options, motif_scan_options) != NO_ERROR) {
    die("Error processing command line options: option name too long.\n");
  }

  while (TRUE) { 
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } 
    else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }
    if (strcmp(option_name, "bgfile") == 0){
      bg_filename = option_value;
    }
    else if (strcmp(option_name, "max-seq-length") == 0) {
      max_seq_length = atoi(option_value);
    } 
    else if (strcmp(option_name, "motif") == 0){
      if (selected_motifs == NULL) {
        selected_motifs = new_string_list();
      }
      add_string(option_value, selected_motifs);
    } 
    else if (strcmp(option_name, "motif-pseudo") == 0){
      pseudocount = atof(option_value);
    } 
    else if (strcmp(option_name, "norc") == 0){
      scan_both_strands = FALSE;
    } 
    else if (strcmp(option_name, "o") == 0){
      // Set output directory with no clobber
      output_dirname = option_value;
      allow_clobber = FALSE;
    } 
    else if (strcmp(option_name, "oc") == 0){
      // Set output directory with clobber
      output_dirname = option_value;
      allow_clobber = TRUE;
    } 
    else if (strcmp(option_name, "output-pthresh") == 0){
      pthresh = atof(option_value);
      qthresh = 1.0;
    } 
    else if (strcmp(option_name, "output-qthresh") == 0){
      qthresh = atof(option_value);
      pthresh = 1.0;
    } 
    else if (strcmp(option_name, "text") == 0){
      text_only = TRUE;
    } 
    else if (strcmp(option_name, "verbosity") == 0){
      verbosity = atoi(option_value);
    }
  }

  // Must have sequence and motif file names
  if (argc != option_index + 2) {
    fprintf(stderr, usage);
    exit(EXIT_FAILURE);
  } 
  char* meme_filename = argv[option_index];
  option_index++;
  char* fasta_filename = argv[option_index];
  option_index++;

  // Create cisml data structure for recording results
  CISML_T *cisml = allocate_cisml("fimo", meme_filename, fasta_filename);
  set_cisml_site_pvalue_cutoff(cisml, pthresh);
  set_cisml_site_qvalue_cutoff(cisml, qthresh);

  /**********************************************
   * Read the motifs.
   **********************************************/
  int num_motifs = 0; 
  int num_motifs_after_rc = 0; 
  MOTIF_T motifs[2 * MAX_MOTIFS];
  STRING_LIST_T* motif_occurrences = NULL;
  BOOLEAN_T has_reverse_strand;
  ARRAY_T* bg_freqs = NULL;
  BOOLEAN_T read_file = FALSE;

  // Assume MEME XML file.
  read_file = read_meme_xml_file(
     meme_filename,
     bg_filename,
     pseudocount,
     &num_motifs,
     motifs,
     &motif_occurrences,
     &has_reverse_strand,
     &bg_freqs
  );

  if (read_file == FALSE) {
    // Wasn't MEME XML, try old file format.
    read_meme_file(
       meme_filename,
       bg_filename,
       pseudocount,
       &num_motifs,
       motifs,
       &motif_occurrences,
       &has_reverse_strand,
       &bg_freqs
    );
  }

  // We don't make use of the motif occurence data
  free_string_list(motif_occurrences);

  // If motifs use protein alphabet we will not scan both  strands
  ALPH_T alphabet_type = which_alphabet();
  if (alphabet_type == PROTEIN_ALPH) {
    scan_both_strands = FALSE;
  } 
  
  if (scan_both_strands == TRUE) {
    // Set up hash tables for computing reverse complement
    setup_hash_alph(DNAB);
    setalph(0);
    // Correct background by averaging on freq. for both strands.
    average_freq_with_complement(bg_freqs);
    int alph_size = get_alph_size(ALPH_SIZE);
    normalize_subarray(0, alph_size, 0.0, bg_freqs);
    fill_in_ambiguous_chars(FALSE, bg_freqs);
    // Make reverse complement motifs.
    num_motifs_after_rc = num_motifs;
    add_reverse_complements(&num_motifs_after_rc, motifs);
    assert(num_motifs_after_rc == (2 * num_motifs));
  }

  /**************************************************************
   * Score each of the sites for each of the selected motifs.
   **************************************************************/
  int motif_index;
  for (motif_index = 0; motif_index < num_motifs; motif_index++) {

    MOTIF_T* motif = &(motifs[motif_index]);
    char* motif_id = get_motif_id(motif);
    char* bare_motif_id = motif_id;
    // Drop the strand indicator from the motif id
    if (*bare_motif_id == '+' || *bare_motif_id == '-') {
      bare_motif_id++;
    }

    // Create cisml pattern and add to cisml record
    PATTERN_T *pattern = allocate_pattern(bare_motif_id, bare_motif_id);
    add_cisml_pattern(cisml, pattern);

    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(
        stderr, 
        "Using motif %s of width %d.\n",
        motif_id,
        motif->length
        );
    }

    // Open the FASTA file for reading.
    FILE* fasta_file = NULL;
    if (open_file(fasta_filename, "r", FALSE, "FASTA", 
      "sequences", &fasta_file) == 0) {
      die("Couldn't open the file %s.\n", fasta_filename);
    }

    if ((selected_motifs == NULL || get_num_strings(selected_motifs) == 0)
      || (have_string(bare_motif_id, selected_motifs) == TRUE)) {

      // Read the FASTA file one sequence at a time.
      SEQ_T* sequence = NULL;
      while (read_one_fasta(fasta_file, max_seq_length, &sequence)) {

        seq_name = get_seq_name(sequence);

        // Create a scanned_sequence record and record it in pattern.
        SCANNED_SEQUENCE_T *scanned_seq = 
          allocate_scanned_sequence(seq_name, seq_name, pattern);
        set_scanned_sequence_length(scanned_seq, get_seq_length(sequence));
        
        if (verbosity >= HIGH_VERBOSE) {
          fprintf(stderr, "Scoring forward strand.\n");
        }

        // Score the forward strand.
        fimo_score_sequence(
          sequence,
          motif, 
          bg_freqs, 
          motif_name, 
          seq_name, 
          pthresh,
          scanned_seq
        );

        // Score the reverse strand.
        if (scan_both_strands) {
          if (verbosity >= HIGH_VERBOSE) {
            fprintf(stderr, "Scoring reverse strand.\n");
          }
          MOTIF_T* rev_motif = &(motifs[num_motifs + motif_index]);
          fimo_score_sequence(
            sequence,
            rev_motif, 
            bg_freqs, 
            motif_name, 
            seq_name, 
            pthresh,
            scanned_seq
          );
        }

        free_seq(sequence);
         
      }  // All sequences parsed

      fclose(fasta_file);

      // If q-value option "on" replace p-values by q-values in site list.
      if (compute_qvalues) {
        pattern_calculate_qvalues(pattern);
      }

    } 
    else {
      if (verbosity >= NORMAL_VERBOSE) {
        fprintf(stderr, "Skipping motif %s.\n", motif_id);
      }
    }


  } // All motifs parsed

  // Write out results
  if (text_only) {
    output_only_text_results(cisml);
  }
  else {
    output_full_results(cisml, output_dirname, allow_clobber);
  }

  // Clean up.
  for (motif_index = 0; motif_index < num_motifs; motif_index++) {
    free_motif(&motifs[motif_index]);
    if (scan_both_strands) {
      free_motif(&motifs[num_motifs + motif_index]);
    }
  }
  free_array(bg_freqs);
  free_string_list(selected_motifs);
  free_cisml(cisml);

  return(0);
}

