#ifndef PSSM_H
#define PSSM_H
#include "mhmm-state.h"

//
// Range of integral score values for a PSSM column.
//
#define PSSM_RANGE 100

//
// Macros to convert between scaled score and raw score.
//
#define scaled_to_raw(x,w,scale,offset) (((x)/(scale)) + ((w)*(offset)))
#define raw_to_scaled(x,w,scale,offset) (nint(((x) - ((w)*offset)) * (scale)))

/************************************************************************
 * Given a model in log format:
 * 	1) Create a PSSM for each motif and attach it to the first state
 *	   of that motif.
 *	2) Create a PSSM for the reverse complement of each motif and
 *     	   attach it to the first state of that motif, if requested.
 *	3) Create a p-value lookup table for match scores to each motif,
 *	   and attach it to the first state of that motif.
 *
 * The PSSM is scaled so that match scores to the motif are in the 
 * range [0..w*range], where w is the width of the motif.
 *
 * If hash_n > 1, the PSSM is hashed to combine hash_n letters into 1
 * for speed in computing match scores.
 *
 * Also, set up the list of "hot" states for faster DP in motif-scoring mode.
 *
 * The PSSM(s) and p-value lookup table are allocated here and must be freed 
 * by the caller.
 *
 *************************************************************************/
void set_up_pssms_and_pvalues (
  BOOLEAN_T motif_scoring,		// Motif scoring?
  double p_threshold,			// Scale/offset PSSM and create table if > 0
  BOOLEAN_T use_both_strands,           // Compute PSSM for negative strand, too?
  BOOLEAN_T allow_weak_motifs,		// Allow motifs with min p-value < p_threshold?
  MHMM_T*  the_hmm                      // The HMM.
);

/**************************************************************************
 * Compute a matrix, for a given sequence, of motif scores w.r.t. each
 * sequence position.  The matrix may be pre-allocated, and it will
 * grow if necessary.
 **************************************************************************/
void compute_motif_score_matrix
  (BOOLEAN_T  use_pvalues,	// Returns scores as p-values, not log-odds.
   double     p_threshold,	// Divide p-values by this.
   int*       int_sequence,
   int        seq_length,
   MHMM_T*    the_hmm,
   MATRIX_T** motif_score_matrix);

/**************************************************************************
*	scale_pssm
*
*	Scale and round the scores in a PSSM so that they
*	are in the range [0...range].
*
*	Returns the scaled PSSM. 
*
**************************************************************************/
void scale_pssm(
  MATRIX_T *pssm,			// The PSSM.
  int range				// The desired range.
);

/**************************************************************************
*
*       get_pv_lookup_pos_dep
*
*       Create a lookup table for motif scores where the background
*       model is position dependent and has the same width as the
*       motif.
*
*       Returns an array of pdf values:
*               pv[x] = Pr(score >= x)
*       for 0<=x<=range*w.
*
*       Assumes:
*               1) motif scores are non-negative, integral
*               2) background model is 0-order Markov
*
**************************************************************************/
ARRAY_T* get_pv_lookup_pos_dep(
  MATRIX_T* pssm,               // The PSSM.
  int range,                    // Maximum score in PSSM.
  int alen,                     // Alphabet size.
  MATRIX_T* background_pssm     // The background model PSSM.
);

/**************************************************************************
*
*	get_pv_lookup
*
*	Create a lookup table for motif scores.
*
*	Returns an array of pdf values:
*		pv[x] = Pr(score >= x)
*	for 0<=x<=range*w.
*
*	Assumes:
*		1) motif scores are non-negative, integral
*		2) background model is 0-order Markov
*
**************************************************************************/
ARRAY_T* get_pv_lookup(
  MATRIX_T* pssm,		// The PSSM.
  int range,			// Maximum score in PSSM.
  int alen,		        // Alphabet size.
  ARRAY_T* background		// The background model.
);

double get_unscaled_score(
  double score,
  int w
);
#endif
